//==========================================================================
//
//      if_ae3069usb.c
//
//      Ethernet device driver for AE-3069USB board
//      ( SL811HS + ADM8511 )
//
//==========================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//####BSDCOPYRIGHTBEGIN####
//
// -------------------------------------------
//
// Portions of this software may have been derived from OpenBSD or other sources,
// and are covered by the appropriate copyright disclaimers included herein.
//
// -------------------------------------------
//
//####BSDCOPYRIGHTEND####
//==========================================================================

#include <pkgconf/system.h>
#include <cyg/infra/cyg_type.h>
#include <cyg/hal/hal_arch.h>
#include <cyg/infra/diag.h>
#include <cyg/hal/hal_io.h>
#include <cyg/hal/plf_intr.h>
#include <cyg/hal/hal_endian.h>
#include <pkgconf/devs_eth_h8300_ae3069usb.h>

#define DEBUG 1

#if DEBUG > 0
#define DEBUG_PRINT(...) diag_printf(__VA_ARGS__)
#define DUMP_BUFFER(ptr, len) diag_dump_buf_with_offset(ptr, len, ptr);
#else
#define DEBUG_PRINT(...)
#define DUMP_BUFFER(ptr, len)
#endif
#if DEBUG > 1
#define DEBUG_PRINT2(...) diag_printf(__VA_ARGS__)
#else
#define DEBUG_PRINT2(...)
#endif

#if defined(CYGPKG_KERNEL)
#include <cyg/kernel/kapi.h>
#define WAIT_COMPLETE() do { sl811_progress(); cyg_thread_delay(1); } while(0)
#else
#define WAIT_COMPLETE() sl811_progress()
#endif

/* ====== SH811HS driver part ====== */

// SL811HS register offset
#define USBACONTROL 0x00
#define USBAADDRESS 0x01
#define USBAXFERLEN 0x02
#define USBASTATUS  0x03
#define USBACOUNTER 0x04
#define CTRLREG    0x05
#define IER        0x06
#define USBBCONTROL 0x08
#define USBBADDRESS 0x09
#define USBBXFERLEN 0x0a
#define USBBSTATUS  0x0b
#define USBBCOUNTER 0x0c
#define STATUS     0x0d
#define DATASET    0x0e
#define CSOFCNT    0x0f
#define DATA_START 0x10

// SL811HS PID TYPE
#define PID_OUT   0x10
#define PID_IN    0x90
#define PID_SETUP 0xd0

#define DESC_DEVICE 0x01
struct usb_dev_desc {
    cyg_uint8  bLength;
    cyg_uint8  bDescriptorType;
    cyg_uint16 bcdUSB;
    cyg_uint8  bDeviceClass;
    cyg_uint8  bDeviceSubClass;
    cyg_uint8  bDeviceProtocol;
    cyg_uint8  bMaxPacketSize0;
    cyg_uint16 idVendor;
    cyg_uint16 idProduct;
    cyg_uint16 bcdDevice;
    cyg_uint8  iManufacturer;
    cyg_uint8  iProduct;
    cyg_uint8  iSerialNumber;
    cyg_uint8  bNumConfigurations;
} __attribute__((aligned(1),packed));

#define DESC_CONFIG 0x02
struct usb_conf_desc {
    cyg_uint8  bLength;
    cyg_uint8  bDescriptorType;
    cyg_uint16 wTotalLength;
    cyg_uint8  bNumInterfaces;
    cyg_uint8  bConfigurationValue;
    cyg_uint8  iConfiguration;
    cyg_uint8  bmAttribute;
    cyg_uint8  bMaxPower;
} __attribute__((aligned(1),packed));

#define DESC_INTERFACE 0x04
struct usb_intf_desc {
    cyg_uint8  bLength;
    cyg_uint8  bDescriptorType;
    cyg_uint8  iInterfaceNumber;
    cyg_uint8  bAlternateSetting;
    cyg_uint8  bNumendpoints;
    cyg_uint8  bInterfaceClass;
    cyg_uint8  bInterfaceSubClass;
    cyg_uint8  bInterfaceProtocol;
    cyg_uint8  iInterface;
} __attribute__((aligned(1),packed));

#define DESC_ENDP 0x05
struct usb_endp_desc {
    cyg_uint8  bLength;
    cyg_uint8  bDescriptorType;
    cyg_uint8  bEndpointAddress;
    cyg_uint8  bmAttribute;
    cyg_uint16 wMaxPacketSize;
    cyg_uint8  bInterval;
} __attribute__((aligned(1),packed));

union usb_desc {
    struct usb_dev_desc  dev       __attribute__((aligned(1)));
    struct usb_conf_desc config    __attribute__((aligned(1)));
    struct usb_intf_desc interface __attribute__((aligned(1)));
    struct usb_endp_desc endp      __attribute__((aligned(1)));
};

struct endpoint {
    cyg_uint16 type;
    short length;
    cyg_uint16 data_state;
};

struct usb_setup {
    cyg_uint8  bRequestType;
    cyg_uint8  bRequest;
    cyg_uint16 wValue;
    cyg_uint16 wIndex;
    cyg_uint16 wSize;
} __attribute__((aligned(1),packed));

#define TOLE16(x) ((x & 0xff00)>>8 | (x & 0x00ff)<<8)
#define TOCPU16(x) ((x & 0xff00)>>8 | (x & 0x00ff)<<8)

#define USB_CLEAR_DEVICE_FEATURE 1
#define USB_SET_ADDRESS       5
#define USB_GET_DESCRIPTOR    6
#define USB_SET_CONFIGURATION 9 

#define REQUEST_OUTPUT 0x00
#define REQUEST_INPUT  0x80

#define EP_MAX 16

static struct endpoint ep[EP_MAX];
static cyg_uint8 config_buffer[256];

enum t_state {t_idle, t_start, t_write, t_read, t_complete, t_end};

static struct {
    cyg_uint8 pid;
    cyg_uint8 addr;
    struct endpoint *ep;
    cyg_uint8 *data;
    int length;
    int status;
    enum t_state transaction;
    int rlength;
    int total;
} sl_req;

static __inline__ void sl811out(cyg_uint8 r, cyg_uint8 d)
{
    HAL_WRITE_UINT8(CYGHWR_H8300_AE3069USB_SL811HS_BASE  , r);
    HAL_WRITE_UINT8(CYGHWR_H8300_AE3069USB_SL811HS_BASE+2, d);
}

static __inline__ cyg_uint8 sl811in(cyg_uint8 r)
{
    cyg_uint8 d;
    HAL_WRITE_UINT8(CYGHWR_H8300_AE3069USB_SL811HS_BASE , r);
    HAL_READ_UINT8(CYGHWR_H8300_AE3069USB_SL811HS_BASE+2, d);
    return d;
}

static __inline__ void sl811outs(cyg_uint8 r, cyg_uint8 *d, int len)
{
    HAL_WRITE_UINT8(CYGHWR_H8300_AE3069USB_SL811HS_BASE  , r);
    for (; len > 0; --len, d++) {
	HAL_WRITE_UINT8(CYGHWR_H8300_AE3069USB_SL811HS_BASE+2, *d);
	HAL_DELAY_US(10);
    }
}

static __inline__ void sl811ins(cyg_uint8 r, cyg_uint8 *d, int len)
{
    HAL_WRITE_UINT8(CYGHWR_H8300_AE3069USB_SL811HS_BASE , r);
    for (; len > 0; --len, d++) {
	HAL_READ_UINT8(CYGHWR_H8300_AE3069USB_SL811HS_BASE+2, *d);
	HAL_DELAY_US(10);
    }
}

static int sl811hs_reset(void)
{
   int s;

    for (s = 0; s < EP_MAX; s++) {
	ep[s].type       = -1;
	ep[s].length     = -1;
	ep[s].data_state = 0;
    }
    ep[s].type         = 0;
    sl_req.pid         = 0;
    sl_req.transaction = t_idle;

    /* reset target */
    sl811out(CTRLREG,    0x08);
#if defined(CYGPKG_KERNEL)
	cyg_thread_delay(25);
#else
	HAL_DELAY_US(250*1000);
#endif
    sl811out(CTRLREG,    0x00);

    sl811out(CSOFCNT,    0xae);
    for (s = 0; s < 100; s++)
    sl811out(STATUS,     0xff);

    sl811out(USBBXFERLEN, 0x00);
    sl811out(USBBSTATUS,  0x50);
    sl811out(USBBCOUNTER, 0x00);
    sl811out(DATASET,    0xe0);
    sl811out(CSOFCNT,    0xae);
    sl811out(CTRLREG,    0x01);
    sl811out(USBBCONTROL, 0x01);
#if defined(CYGPKG_KERNEL)
	cyg_thread_delay(1);
#else
	HAL_DELAY_US(10*1000);
#endif
    s = sl811in(STATUS);
    DEBUG_PRINT2("STATUS=%02x\n",s);
    if ((s & 0x40) != 0) {
	diag_printf("USB device not present\n");
	return 0;
    }
    if ((s & 0x80) == 0) {
	DEBUG_PRINT("USB Low speed device detect\n");
       	sl811out(CSOFCNT,   0xee);
	sl811out(CTRLREG,   0x21);
	sl811out(USBBCONTROL,0x01);
    } else {
	DEBUG_PRINT("USB High speed device detect\n");
       	sl811out(CSOFCNT,   0xae);
	sl811out(CTRLREG,   0x01);
	sl811out(USBBCONTROL,0x01);
    }
    sl811out(IER, 0x00);
    return 1;
}

static void sl811setup(cyg_uint8 pid, cyg_uint8 addr, cyg_uint8 len)
{
    sl811out(USBAADDRESS, DATA_START);
    sl811out(USBAXFERLEN, len);
    sl811out(USBASTATUS,  pid);
    sl811out(USBACOUNTER, addr);
}

#if DEBUG > 2
static void print_request(void)
{
    diag_printf("PID=%02x, ADDR=%02x, LEN=%04x, STATUS=%02x\n", sl_req.pid, sl_req.addr, sl_req.length, sl_req.status);
}
#endif

static void sl811_progress(void)
{
    int wlen;
    cyg_uint8 sl_status;

    sl_status = sl811in(STATUS);
    // busy check
    if (sl_req.transaction != t_start) {
	if (sl_req.transaction == t_idle || (sl_status & 0x01)==0) {
	    sl811out(STATUS, sl_status);
	    return ;
	}

	sl_req.status = sl811in(USBASTATUS);
	sl_req.ep->data_state ^= 0x40;
	if (sl_req.transaction == t_complete || (sl_req.status & 0xe6)) {
	    sl_req.pid = 0;
	    sl_req.transaction = t_end;
	    sl811out(STATUS, sl_status);
	    return;
	}
    }

    switch (sl_req.pid & 0xf0) {
    case PID_OUT:
	wlen = sl_req.length;
	if (sl_req.ep->length >= 0 && sl_req.ep->length < wlen)
	    wlen = sl_req.ep->length;
	sl811setup(sl_req.pid, sl_req.addr, wlen);
	DEBUG_PRINT2("write ptr=%p, size=%d, state=%02x\n",
		     sl_req.data, wlen, sl_req.ep->data_state);
	sl811outs(DATA_START, sl_req.data, wlen);
#if DEBUG > 2
	print_request();
	DUMP_BUFFER(sl_req.data, wlen);
#endif
	sl811out(USBACONTROL, 0x07 | sl_req.ep->data_state);
	sl_req.data   += wlen;
	sl_req.length -= wlen;
	sl_req.transaction = (sl_req.length == 0)?t_complete:t_write;
	break ;
    case PID_IN:
	if (sl_req.transaction == t_read) {
	    sl_req.rlength -= sl811in(USBACOUNTER);
	    if (sl_req.rlength > 0) {
		DEBUG_PRINT2("read ptr=%p, size=%d, state=%02x \n",
			     sl_req.data, sl_req.rlength, sl_req.ep->data_state);
		sl811ins(DATA_START, sl_req.data, sl_req.rlength);
#if DEBUG > 2
		DUMP_BUFFER(sl_req.data, sl_req.rlength);
#endif
		sl_req.data   += sl_req.rlength;
		sl_req.total  += sl_req.rlength;
		sl_req.length -= sl_req.rlength;
		if (sl_req.rlength < sl_req.ep->length) {
		    sl_req.pid = 0;
		    sl_req.transaction = t_end;
		    break ; 
		}
	    } else {
		DEBUG_PRINT2("read size = 0\n");
		sl_req.pid = 0;
		sl_req.transaction = t_end;
		break ; 
	    }
	}
#if DEBUG > 2
	print_request();
#endif
	sl_req.rlength = sl_req.length;
	if (sl_req.ep->length >= 0 && sl_req.ep->length < sl_req.rlength)
	    sl_req.rlength = sl_req.ep->length;
	sl811setup(sl_req.pid, sl_req.addr, sl_req.rlength);
	sl811out(USBACONTROL, 0x03 | sl_req.ep->data_state);
	sl_req.transaction = t_read;
	break ;
    case PID_SETUP:
#if DEBUG > 2
	DEBUG_PRINT2("setup\n");
	DUMP_BUFFER(sl_req.data, 8);
#endif
	sl811setup(sl_req.pid, sl_req.addr, 8);
	sl811outs(DATA_START, sl_req.data, 8);
	sl811out(USBACONTROL, 0x07);
	sl_req.ep->data_state = 0x00;
	sl_req.transaction = t_complete;
	break ;
    }
    sl811out(STATUS, sl_status);
}

static __inline__ int sl811_status(void)
{
    switch (sl_req.transaction) {
    case t_idle: return 0;
    case t_end: 
	sl_req.transaction = t_idle;
	return sl_req.status & 0xff;
    default:
	return -1;
    }
}

static __inline__ int sl811_rxlen(void)
{
    return sl_req.total;
}

static int sl811_write_data(cyg_uint8 endp, cyg_uint8 addr, cyg_uint8 *data, int len)
{
    if (sl_req.transaction != t_idle)
	return 0;

    sl_req.pid         = PID_OUT | (endp & 0x0f);
    sl_req.addr        = addr;
    sl_req.ep          = &ep[endp];
    sl_req.data        = data;
    sl_req.length      = len;
    sl_req.transaction = t_start;

    return 1;
}

static int sl811_read_data(cyg_uint8 endp, cyg_uint8 addr, cyg_uint8 *data, int len)
{
    if (sl_req.transaction != t_idle)
	return 0;

    sl_req.pid         = PID_IN | (endp & 0x0f);
    sl_req.addr        = addr;
    sl_req.ep          = &ep[endp];
    sl_req.data        = data;
    sl_req.length      = len;
    sl_req.total       = 0;
    sl_req.transaction = t_start;

    return 1;
}

static int send_setup(cyg_uint8 addr, struct usb_setup *setup)
{
    if (sl_req.transaction != t_idle)
	return 0;

    sl_req.pid         = PID_SETUP;
    sl_req.addr        = addr;
    sl_req.ep          = &ep[0];
    sl_req.data        = (cyg_uint8 *)setup;
    sl_req.transaction = t_start;

    return 1;
}

static bool get_descriptor(cyg_uint8 addr,int type, union usb_desc *desc, int len)
{
    struct usb_setup setup;
    int s;
    cyg_uint8 *desc_p;
    
    if (ep[0].length < 0) len = 8;
    setup.bRequestType  = REQUEST_INPUT;
    setup.bRequest      = USB_GET_DESCRIPTOR;
    setup.wValue        = TOLE16(type << 8);
    setup.wIndex        = TOLE16(0);
    setup.wSize         = TOLE16(len);
    send_setup(addr, &setup);
    while ((s = sl811_status()) == -1)
	WAIT_COMPLETE();
    if (s & 0xe6)
	return false;
    desc_p = (cyg_uint8 *)desc;
    sl811_read_data(0, addr, desc_p, len);
    while (sl811_status() == -1)
	WAIT_COMPLETE();
    sl811_write_data(0,addr, desc_p, 0);
    while (sl811_status() == -1)
	WAIT_COMPLETE();
    return true;
}

static bool set_address(cyg_uint8 old, cyg_uint8 new)
{
    struct usb_setup setup;
    int s;
    setup.bRequestType  = REQUEST_OUTPUT;
    setup.bRequest      = USB_SET_ADDRESS;
    setup.wValue        = TOLE16(new);
    setup.wIndex        = TOLE16(0);
    setup.wSize         = TOLE16(0);
    send_setup(old, &setup);
    while ((s = sl811_status()) == -1)
	WAIT_COMPLETE();
    if (s & 0xe6)
	return false;
    sl811_read_data(0, old, (cyg_uint8 *)&setup, 0);
    while (sl811_status() == -1)
	WAIT_COMPLETE();
    return true;
}

static bool clear_device_feature(cyg_uint8 addr,cyg_uint16 feature)
{
    struct usb_setup setup;
    int s;

    setup.bRequestType  = REQUEST_OUTPUT;
    setup.bRequest      = USB_CLEAR_DEVICE_FEATURE;
    setup.wValue        = TOLE16(feature);
    setup.wIndex        = TOLE16(0);
    setup.wSize         = TOLE16(0);
    send_setup(addr, &setup);
    while ((s = sl811_status()) == -1)
	WAIT_COMPLETE();
    if (s & 0xe6)
	return false;
    sl811_read_data(0, addr, (cyg_uint8 *)&setup, 0);
    while (sl811_status() == -1)
	WAIT_COMPLETE();
    return true;
}

static bool set_configuration(cyg_uint8 addr,cyg_uint16 config)
{
    struct usb_setup setup;
    int s;

    setup.bRequestType  = REQUEST_OUTPUT;
    setup.bRequest      = USB_SET_CONFIGURATION;
    setup.wValue        = TOLE16(config);
    setup.wIndex        = TOLE16(0);
    setup.wSize         = TOLE16(0);
    send_setup(addr, &setup);
    while ((s = sl811_status()) == -1)
	WAIT_COMPLETE();
    if (s & 0xe6)
	return false;
    sl811_read_data(0, addr, (cyg_uint8 *)&setup, 0);
    while (sl811_status() == -1)
	WAIT_COMPLETE();
    return true;
}

static bool parse_config(cyg_uint8 *config_buffer, int config_len, struct endpoint *ep)
{
    union usb_desc *desc;
    int num_ep = 0;

    desc  = (union usb_desc *)config_buffer;
    while(config_len > 0) {
#if DEBUG > 2
	DUMP_BUFFER(desc, desc->config.bLength);
#endif
	switch(desc->config.bDescriptorType) {
	case DESC_CONFIG:
	    break ;
	case DESC_INTERFACE:
	    num_ep = desc->interface.bNumendpoints;
	    DEBUG_PRINT2("bNumendpoints = %d\n",num_ep);
	    if (num_ep >= EP_MAX) {
		diag_printf("Illigal Configuration\n");
		return false;
	    }
	    break ;
	case DESC_ENDP:
	    if (num_ep < (desc->endp.bEndpointAddress & 0x7f)) {
		diag_printf("Illigal Configuration\n");
		return false;
	    }
	    ep[desc->endp.bEndpointAddress & 0x7f].type    = ((desc->endp.bEndpointAddress & 0x80) << 1) |
		                                             desc->endp.bmAttribute;
	    ep[desc->endp.bEndpointAddress & 0x7f].length  = *(cyg_uint8 *)&(desc->endp.wMaxPacketSize) |
		                                             *((cyg_uint8 *)&(desc->endp.wMaxPacketSize) + 1) << 8;
	    DEBUG_PRINT2("EP%d: type = %02x, len = %d\n",
			desc->endp.bEndpointAddress & 0x7f,
			ep[desc->endp.bEndpointAddress & 0x7f].type,
			ep[desc->endp.bEndpointAddress & 0x7f].length);
	    break ;
	}
	config_len -= desc->config.bLength;
	desc = (union usb_desc *)((cyg_uint8 *)desc + desc->config.bLength);
    }
    return true;
}

static bool init_target(cyg_uint16 vender, cyg_uint16 product, cyg_uint8 addr)
{
    union usb_desc desc;
    int config_len;
    int target_config;

    if (!get_descriptor(0, DESC_DEVICE, &desc, 8))
	return false;
    ep[0].length = desc.dev.bMaxPacketSize0;
    DEBUG_PRINT2("endpoint0 size=%d\n",ep[0].length);

#if DEBUG > 2
    DUMP_BUFFER(&desc, 8);
#endif

    if (!set_address(0, addr))
	return false;
    if (!get_descriptor(addr, DESC_DEVICE, &desc, sizeof(desc.dev)))
	return false;
#if DEBUG > 2
    DUMP_BUFFER(&desc, sizeof(desc.dev));
#endif

    DEBUG_PRINT2("\nDeviceID %04x:%04x\n",TOCPU16(desc.dev.idVendor),TOCPU16(desc.dev.idProduct));
    if (TOCPU16(desc.dev.idVendor)  != vender|| TOCPU16(desc.dev.idProduct) != product)
	/* unknown device */
	return false;
    if (!get_descriptor(addr, DESC_CONFIG, &desc, sizeof(desc.config)))
	return false;
#if DEBUG > 1
    DUMP_BUFFER(&desc, sizeof(desc.config));
#endif
    DEBUG_PRINT2("wTotalLength = %d\n",TOCPU16(desc.config.wTotalLength));
    DEBUG_PRINT2("bNumInterfaces = %d\n",desc.config.bNumInterfaces);
    DEBUG_PRINT2("bConfigurationValue = %d\n",desc.config.bConfigurationValue);
    target_config = desc.config.bConfigurationValue;
    config_len = TOCPU16(desc.config.wTotalLength);
    if (sizeof(config_buffer) < config_len) {
	diag_printf("config desc too long\n");
	return false;
    }
    if (!get_descriptor(addr, DESC_CONFIG, (union usb_desc *)config_buffer, config_len))
	return false;
#if DEBUG > 2
    DUMP_BUFFER(config_buffer, config_len);
#endif
    if (!parse_config(config_buffer, config_len, ep))
	return false;
    if (!set_configuration(addr, target_config))
	return false;
    if (!clear_device_feature(addr, 1))
	return false;

    return true;
}

/* ==== SH811HS driver part end ==== */

/* ====== pegasus driver part ====== */
#include <cyg/io/eth/eth_drv.h>
#include <cyg/io/eth/netdev.h>
#include <cyg/hal/drv_api.h>

#define	PEGASUS_READ	    0xc0
#define	PEGASUS_WRITE	    0x40
#define	PEGASUS_CMD_GETREGS 0xf0
#define	PEGASUS_CMD_SETREGS 0xf1

#define PEGASUS_ADDR 2

#define REG_ETHERCTRL0 0x00
#define REG_ETHERCTRL1 0x01
#define REG_ETHERCTRL2 0x02
#define REG_ETHERID0   0x10
#define REG_PHYADDR    0x25
#define REG_PHYDATA    0x26
#define REG_PHYDATAL   0x26
#define REG_PHYDATAH   0x27
#define REG_PHYCTRL    0x28
#define REG_WAKEUPSTS  0x7a
#define REG_IPHYCTRL   0x7b
#define REG_GPIO0      0x7e
#define REG_GPIO1      0x7f

#define PHY_READ       0x40

#define MAX_PHY_INDEX  31
#define PHY_DEFAULT_IDX 1

#define MII_BMCR            0x00
#define MII_BMSR            0x01
#define MII_LPA             0x05

#define BMSR_MEDIA    0x7808

#define LINK_100M     0x0180
#define LINK_FDX      0x0140

#define RESET_TIMEOUT 100
#define PHY_TIMEOUT   1000
#define LINKCHECK_INTERVAL 10000

#define EP_RX 1
#define EP_TX 2
#define EP_STATUS 3

enum usb_operation {rx_req, tx_start, tx_now, tx_fin, stat_req};

typedef struct pegasus_priv_data {
    cyg_uint8 addr;
    cyg_uint8 phy;
    cyg_uint8 linkst;
    cyg_uint8 esa[6];
    bool hardwired_esa;
    enum usb_operation usbop;
    int txlen;
    int linkchk_time;
    unsigned int key;
    cyg_vector_t interrupt;
    cyg_handle_t  interrupt_handle;
    cyg_interrupt interrupt_object;
    cyg_uint8 tx_buffer[1514+2];
    cyg_uint8 rx_buffer[1514+8];
} pegasus_priv_data_t;

static struct pegasus_priv_data pegasus_eth0_priv_data = {
    addr: CYGNUM_H8300_AE3069USB_ETH_ADDR,
    phy: -1,
    interrupt: CYGHWR_H8300_AE3069USB_SL811HS_INT,
#if defined(CYGSEM_DEVS_ETH_H8300_AE3069USB_ETH0_SET_ESA)
    esa: CYGDAT_DEVS_ETH_H8300_AE3069USB_ETH0_ESA,
    hardwired_esa: true,
#else
    hardwired_esa: false,
#endif
};

ETH_DRV_SC(adm8511_sc,
           &pegasus_eth0_priv_data, // Driver specific datadesc_p - (cyg_uint8 *)desc;
           CYGDAT_DEVS_ETH_H8300_AE3069USB_ETH0_NAME,
           adm8511_start,
           adm8511_stop,
           adm8511_control,
           adm8511_can_send,
           adm8511_send,
           adm8511_recv,
           adm8511_deliver,     // "pseudoDSR" called from fast net thread
           adm8511_poll,        // poll function, encapsulates ISR and DSR
           adm8511_int_vector);

NETDEVTAB_ENTRY(adm8511_netdev, 
                "adm8511_" CYGDAT_DEVS_ETH_H8300_AE3069USB_ETH0_NAME,
                adm8511_init, 
                &adm8511_sc);

static bool pegasus_get_registers(cyg_uint8 addr, cyg_uint16 index, cyg_uint16 size, cyg_uint8 *data)
{
    struct usb_setup setup;
    int s;
    setup.bRequestType  = PEGASUS_READ;
    setup.bRequest      = PEGASUS_CMD_GETREGS;
    setup.wValue        = TOLE16(0);
    setup.wIndex        = TOLE16(index);
    setup.wSize         = TOLE16(size);
    send_setup(addr, &setup);
    while ((s = sl811_status()) == -1)
	WAIT_COMPLETE();
    if (s & 0xe6)
	return false;
    sl811_read_data(0, addr, data, size);
    while (sl811_status() == -1)
	WAIT_COMPLETE();
    sl811_write_data(0, addr, (cyg_uint8 *)&setup, 0);
    while (sl811_status() == -1)
	WAIT_COMPLETE();
    return true;
}

static bool pegasus_set_registers(cyg_uint8 addr, cyg_uint16 index, cyg_uint16 size, cyg_uint8 *data)
{
    struct usb_setup setup;
    int s;

    setup.bRequestType  = PEGASUS_WRITE;
    setup.bRequest      = PEGASUS_CMD_SETREGS;
    setup.wValue        = TOLE16(0);
    setup.wIndex        = TOLE16(index);
    setup.wSize         = TOLE16(size);
    send_setup(addr, &setup);
    while ((s = sl811_status()) == -1)
	WAIT_COMPLETE();
    if ((s & 0xe6))
	return false;
    sl811_write_data(0, addr, data, size);
    while ((s = sl811_status()) == -1)
	WAIT_COMPLETE();
    if ((s & 0xe6))
	return false;
    sl811_read_data(0, addr, (cyg_uint8 *)&setup, 0);
    while (sl811_status() == -1)
	WAIT_COMPLETE();
    return true;
}

static bool pegasus_set_register(cyg_uint16 addr, cyg_uint16 index, cyg_uint8 data)
{
    struct usb_setup setup;
    int s;

    setup.bRequestType  = PEGASUS_WRITE;
    setup.bRequest      = PEGASUS_CMD_SETREGS;
    setup.wValue        = TOLE16(data);
    setup.wIndex        = TOLE16(index);
    setup.wSize         = TOLE16(1);
    send_setup(addr, &setup);
    while ((s = sl811_status()) == -1)
	WAIT_COMPLETE();
    if (s & 0xe6)
	return false;
    sl811_write_data(0, addr, &data, 1);
    while ((s = sl811_status()) == -1)
	WAIT_COMPLETE();
    if ((s & 0xe6))
	return false;
    sl811_read_data(0, addr, (cyg_uint8 *)&setup, 0);
    while (sl811_status() == -1)
	WAIT_COMPLETE();
    return true;
}

static bool reset_mac(cyg_uint8 addr)
{
    int t;
    cyg_uint8 ethctrl1;
    DEBUG_PRINT("reset mac...");
    pegasus_set_register(addr, REG_ETHERCTRL1, 0x08);  // reset mac
    for(t = 0; t < RESET_TIMEOUT; t++) {
#if defined(CYGPKG_KERNEL)
	cyg_thread_delay(1);
#else
	HAL_DELAY_US(1000);
#endif
	pegasus_get_registers(addr, REG_ETHERCTRL1, 1, &ethctrl1);
	if (!(ethctrl1 & 0x08)) {
	    // reset done
	    DEBUG_PRINT("done.\n");
	    pegasus_set_register(addr, REG_GPIO1, 0x26);
	    pegasus_set_register(addr, REG_GPIO0, 0x24);
	    pegasus_set_register(addr, REG_GPIO0, 0x26);
	    break;
	}
    }
    return (t < RESET_TIMEOUT);
}

static bool read_mii(cyg_uint8 addr, cyg_uint8 phy, cyg_uint8 idx, cyg_uint16 *data)
{
    cyg_uint8 regs[4] = {phy, 0, 0, idx};
    cyg_uint16 phy_status;
    int t;

    pegasus_set_register(addr, REG_PHYCTRL, 0);
    pegasus_set_registers(addr, REG_PHYADDR, sizeof(regs), regs);
    if (!pegasus_set_register(addr, REG_PHYCTRL, PHY_READ | idx))
	return false;

    for (t = 0; t < PHY_TIMEOUT; t++) {
	pegasus_get_registers(addr, REG_PHYCTRL, 1, regs);
	if (regs[0] & 0x80)
	    break ;
#if defined(CYGPKG_KERNEL)
	cyg_thread_delay(1);
#else
	HAL_DELAY_US(1000);
#endif
    }
    if (t < PHY_TIMEOUT) {
	pegasus_get_registers(addr, REG_PHYDATA, 2, (cyg_uint8 *)&phy_status);
	*data = CYG_LE16_TO_CPU(phy_status);
	return true;
    }
    return false;
}

static int mii_probe(cyg_uint8 addr)
{
    int phy_idx;
    cyg_uint16 mii_data;

    // internal PHY poweron/reset
    pegasus_set_register(addr, 0x1d, 0);
    pegasus_set_register(addr, REG_IPHYCTRL, 0x01);
#if defined(CYGPKG_KERNEL)
	cyg_thread_delay(1);
#else
	HAL_DELAY_US(10000);
#endif
    pegasus_set_register(addr, REG_IPHYCTRL, 0x02);

    pegasus_set_register(addr, 0x80, 0xc0 );
    pegasus_set_register(addr, 0x83, 0xff );
    pegasus_set_register(addr, 0x84, 0x01 );
    
    pegasus_set_register(addr, 0x81, 2);

    for (phy_idx = 0; phy_idx < MAX_PHY_INDEX; phy_idx++) {
	if (!read_mii(addr, phy_idx, MII_BMSR, &mii_data)) {
	    DEBUG_PRINT("mii timeout\n");
	    continue;
	}
	DEBUG_PRINT("mii(%d) BMSR=%04x\n", phy_idx, mii_data);
	if ((mii_data != 0) && (mii_data != 0xffff) && ((mii_data & BMSR_MEDIA) != 0)) {
	    DEBUG_PRINT("PHY found. index=%d\n", phy_idx);
	    return phy_idx;
	}
    }
    DEBUG_PRINT("PHY probe failed. use defaut index=%d\n",PHY_DEFAULT_IDX);
    return PHY_DEFAULT_IDX;
}

#if defined(CYGINT_IO_ETH_INT_SUPPORT_REQUIRED)
// This ISR is called when the ethernet interrupt occurs
static cyg_uint32 adm8511_isr(cyg_vector_t vector, cyg_addrword_t data)
{
    struct eth_drv_sc *sc = (struct eth_drv_sc *)data;
    pegasus_priv_data_t *priv = (pegasus_priv_data_t *)sc->driver_private;

    cyg_drv_interrupt_mask(priv->interrupt);
    cyg_drv_interrupt_acknowledge(priv->interrupt);
    return (CYG_ISR_HANDLED|CYG_ISR_CALL_DSR);  // Run the DSR
}

static void adm8511_dsr(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t sc)
{
    // This conditioning out is necessary because of explicit calls to this
    // DSR - which would not ever be called in the case of a polled mode
    // usage ie. in RedBoot.

    eth_drv_dsr( vector, count, sc );
}
#endif // CYGINT_IO_ETH_INT_SUPPORT_REQUIRED

#if (CYGHWR_H8300_AE3069USB_ETH_VENDOR == 0) && (CYGHWR_H8300_AE3069USB_ETH_PRODUCT == 0)
#error illigal configuration device vender ID and product ID
#endif

static bool adm8511_init(struct cyg_netdevtab_entry *tab)
{
    struct eth_drv_sc *sc = (struct eth_drv_sc *)tab->device_instance;

    pegasus_priv_data_t *priv = (pegasus_priv_data_t *)sc->driver_private;

    if (!sl811hs_reset())
	return false;

    if (!init_target(CYGHWR_H8300_AE3069USB_ETH_VENDOR, 
		     CYGHWR_H8300_AE3069USB_ETH_PRODUCT, 
		     priv->addr))
	return false;

    DEBUG_PRINT("mac reset.\n");
    if (!reset_mac(priv->addr))
	return false;

    DEBUG_PRINT("phy setup.\n");
    priv->phy = mii_probe(priv->addr);

    if (!priv->hardwired_esa)
	pegasus_get_registers(priv->addr, REG_ETHERID0, sizeof(priv->esa), priv->esa);
    else
	pegasus_set_registers(priv->addr, REG_ETHERID0, sizeof(priv->esa), priv->esa);

    diag_printf("ADM8511 - %s ESA: %02x:%02x:%02x:%02x:%02x:%02x\n",
                (priv->hardwired_esa) ? "static" : "eeprom",
                priv->esa[0],
                priv->esa[1],
                priv->esa[2],
                priv->esa[3],
                priv->esa[4],
                priv->esa[5] );

#if 0
#ifdef CYGINT_IO_ETH_INT_SUPPORT_REQUIRED
    cyg_drv_interrupt_create(
        priv->interrupt,
        0,                         // Priority - unused
        (cyg_addrword_t)sc,        // Data item passed to ISR & DSR
        adm8511_isr,               // ISR
        adm8511_dsr,               // DSR
        &priv->interrupt_handle,   // handle to intr obj
        &priv->interrupt_object ); // space for int obj

    cyg_drv_interrupt_attach(priv->interrupt_handle);
    cyg_drv_interrupt_unmask(priv->interrupt);
#endif
#endif

    // Initialize upper level driver
    (sc->funs->eth_drv->init)(sc, priv->esa);

    return true;
}

static void adm8511_start(struct eth_drv_sc *sc, unsigned char *enaddr, int flags)
{
    pegasus_priv_data_t *priv = (pegasus_priv_data_t *)sc->driver_private;
    cyg_uint16 bmsr;
    cyg_uint16 link;
    cyg_uint8 regs[3] = {0xc9, 0x00, 0x01};

    read_mii(priv->addr, priv->phy, MII_BMSR, &bmsr);
    read_mii(priv->addr, priv->phy, MII_LPA, &link);
    regs[1] |= (link & LINK_FDX)? 0x20:0x00;
    regs[1] |= (link & LINK_100M)?0x10:0x00;
    pegasus_set_registers(priv->addr, REG_ETHERCTRL0, sizeof(regs), regs);
    pegasus_get_registers(priv->addr, REG_WAKEUPSTS, sizeof(priv->linkst), &priv->linkst);
    priv->usbop = rx_req;
    priv->linkchk_time = LINKCHECK_INTERVAL;
}

static void adm8511_stop(struct eth_drv_sc *sc)
{
    pegasus_priv_data_t *priv = (pegasus_priv_data_t *)sc->driver_private;
    cyg_uint8 regs[2] = {0x00, 0x00};

    pegasus_set_registers(priv->addr, REG_ETHERCTRL0, sizeof(regs), regs);
}

static int adm8511_control(struct eth_drv_sc *sc, unsigned long key,
               void *data, int data_len)
{
    pegasus_priv_data_t *priv = (pegasus_priv_data_t *)sc->driver_private;

    switch (key) {
    case ETH_DRV_SET_MAC_ADDRESS:
	if (data_len != 6)
	    return -2;
	pegasus_set_registers(priv->addr, REG_ETHERID0, 6, (cyg_uint8 *)data);
        return 0;
        break;
    default:
        return 1;
        break;
    }
}

static int adm8511_can_send(struct eth_drv_sc *sc)
{
    pegasus_priv_data_t *priv = (pegasus_priv_data_t *)sc->driver_private;

    return (priv->usbop == rx_req) && ((priv->linkst & 1) != 0);
}

#if defined(CYGPKG_REDBOOT)
static int idlecall_count;
#endif

static void adm8511_send(struct eth_drv_sc *sc, struct eth_drv_sg *sg_list, int sg_len, 
			 int total_len, unsigned long key)
{
    pegasus_priv_data_t *priv = (pegasus_priv_data_t *)sc->driver_private;

    int len;
    int idx;
    cyg_uint8 *s,*d;

    DEBUG_PRINT(__FUNCTION__"\n");
    d = priv->tx_buffer + 2;
    // copy sg_list to tx_buffer
    for (idx = 0; idx < sg_len; idx++) {
	s = (cyg_uint8 *)sg_list[idx].buf;
	for (len = 0; len < sg_list[idx].len; len++)
	    *d++ = *s++;
    }
    //padding to shortpacket
    if (total_len < 64)
	for(len = 64 - total_len; len > 0; len--)
	    *d++ = 0;

    len = d - priv->tx_buffer;

    *(cyg_uint16 *)(priv->tx_buffer) = CYG_CPU_TO_LE16(total_len+2);

    //send start
    DEBUG_PRINT(__FUNCTION__ " len=%d\n", total_len);
    DUMP_BUFFER(priv->tx_buffer+2, total_len);
    priv->usbop  = tx_start;
    priv->txlen = len;
    priv->key = key;
    //When interrupt stops, transmit it and start
#if defined(CYGPKG_REDBOOT)
    idlecall_count++;  // Loop transfer disabled
#endif
    adm8511_poll(sc);
}

static void adm8511_recv(struct eth_drv_sc *sc, struct eth_drv_sg *sg_list, int sg_len)
{
    pegasus_priv_data_t *priv = (pegasus_priv_data_t *)sc->driver_private;
    cyg_uint8 *s,*d;
    int idx,len;

    DEBUG_PRINT(__FUNCTION__"\n");
    s = priv->rx_buffer;
    for (idx = 0; idx < sg_len; idx++) {
	d = (cyg_uint8 *)sg_list[idx].buf;
	for (len = 0; len < sg_list[idx].len; len++)
	    *d++ = *s++;
    }
}

static void adm8511_poll(struct eth_drv_sc *sc)
{
    pegasus_priv_data_t *priv = (pegasus_priv_data_t *)sc->driver_private;
    int s;
    int len;
    bool loop_on_complete;
#if defined(CYGPKG_REDBOOT)
    loop_on_complete = (idlecall_count == 0);
    idlecall_count = 0;
#else
    loop_on_complete = false;
#endif
    do {
	sl811_progress();
	s = sl811_status();
	if (s == 0) {
	    // USB idle
	    if (priv->usbop == tx_start) {
		DEBUG_PRINT("Send Packet len=%d\n",priv->txlen);
		if (sl811_write_data(EP_TX, priv->addr, priv->tx_buffer, priv->txlen))
		    priv->usbop = tx_now;
		else {
		    diag_printf("send request error\n");
		    priv->usbop = rx_req;
		}
	    } else if ((priv->linkchk_time == 0) || (priv->linkst & 1) == 0) {
		sl811_read_data(EP_STATUS, priv->addr, priv->rx_buffer, 8);
		priv->usbop = stat_req;
		priv->linkchk_time = LINKCHECK_INTERVAL;
	    } else
		sl811_read_data(EP_RX, priv->addr, priv->rx_buffer, sizeof(priv->rx_buffer));
	} else if(s > 0) {
	    if (s & 0x01) {  // ACK
		switch (priv->usbop) {
		case tx_now:
		    if (priv->txlen % 64 == 0) {
			// Add packet of 0 bytes when transmit packet 
			// of length that can finish dividing by 64
			if (sl811_write_data(EP_TX, priv->addr, priv->tx_buffer, 0)) {
			    priv->usbop =  tx_fin;
			    if(loop_on_complete)
				continue;
			} else {
			    diag_printf("send request error\n");
			    priv->usbop = rx_req;
			}
			break ;
		    }
		case tx_fin:
		    // packet tx complete
		    DEBUG_PRINT("Send complete\n");
		    priv->usbop = rx_req;
		    (sc->funs->eth_drv->tx_done)(sc, priv->key, 0);
		    sl811_read_data(EP_RX, priv->addr, priv->rx_buffer, sizeof(priv->rx_buffer));
		    break;
		case stat_req:
		    priv->linkst = priv->rx_buffer[5];
		    priv->usbop = rx_req;
		    if ((priv->linkst & 1) == 0)
			continue;
		    break;
		default:
		    // packet rx complete
		    DEBUG_PRINT("Receive Packet %02x:%02x:%02x:%02x:%02x:%02x\n",
				priv->rx_buffer[6],priv->rx_buffer[7],priv->rx_buffer[8],
				priv->rx_buffer[9],priv->rx_buffer[10],priv->rx_buffer[11]);
		    len = sl811_rxlen();
		    DUMP_BUFFER(priv->rx_buffer, len);
		    // rx status
		    s = priv->rx_buffer[len - 2];
		    s |= priv->rx_buffer[len - 1] << 8;
		    if ((s & 0x1e) == 0) {
			(sc->funs->eth_drv->recv)(sc, len - 8);
		    }
		    break;
		}
	    } else {
		if (priv->usbop == tx_now) {
		    diag_printf("Tx Error code:%02x\n",s);
		    priv->usbop = rx_req;
		} else if (s != 0x40)
		    // NAK is ignored
		    diag_printf("Rx Error code:%02x\n",s);
	    }
	    loop_on_complete = false;
	}
	priv->linkchk_time--;
    } while(loop_on_complete);
    //DEBUG_PRINT("poll done\n");
}

static int adm8511_int_vector(struct eth_drv_sc *sc)
{
    pegasus_priv_data_t *priv = (pegasus_priv_data_t *)sc->driver_private;
    return priv->interrupt;
}

static void adm8511_deliver(struct eth_drv_sc *sc)
{
    pegasus_priv_data_t *priv = (pegasus_priv_data_t *)sc->driver_private;

    // Service the interrupt:
#if defined(CYGPKG_REDBOOT)
    idlecall_count++;
#endif
    adm8511_poll(sc);
    // Allow interrupts to happen again

    //cyg_drv_interrupt_unmask(priv->interrupt);

}

#if defined(CYGPKG_REDBOOT)
#include <redboot.h>

static void adm8511_redboot_idle(bool is_idle)
{
    if (is_idle) {
	idlecall_count++;
	sl811_progress();
    }
}

RedBoot_idle(adm8511_redboot_idle, RedBoot_IDLE_BEFORE_NETIO);
#endif

/* ==== pegasus driver part end ==== */
