#ifndef CYGONCE_DEVS_FLASH_H8300_PROGS_INL
#define CYGONCE_DEVS_FLASH_H8300_PROGS_INL
//==========================================================================
//
//      h8300progs.inl
//
//      Renesas H8/300 series interlanl flash program type driver
//
//==========================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
// Copyright (C) 2002 Gary Thomas
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//==========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    gthomas
// Contributors: yoshinori sato
// Date:         2004-04-02
// Purpose:      
// Description:  Renesas H8/300 series interanl flash progaming type driver
//
//####DESCRIPTIONEND####
//
//==========================================================================

#include <pkgconf/hal.h>
#include <pkgconf/devs_flash_h8300_progs.h>
#include <cyg/hal/hal_arch.h>
#include <cyg/hal/hal_cache.h>
#include <cyg/hal/hal_misc.h>
#include <cyg/hal/hal_intr.h>
#include <cyg/hal/var_arch.h>
#include <cyg/infra/diag.h>
#include CYGHWR_MEMORY_LAYOUT_H

#define  _FLASH_PRIVATE_
#include <cyg/io/flash.h>

#define CYGNUM_FLASH_BLANK              (1)

//----------------------------------------------------------------------------
// Now that device properties are defined, include magic for defining
// accessor type and constants.
#include <cyg/io/flash_dev.h>

//----------------------------------------------------------------------------
// Information about supported devices
typedef struct flash_dev_info {
    cyg_uint32   block_size;
    cyg_int32    block_count;
    cyg_uint32   device_size;
    cyg_bool     bootblock;
    cyg_uint32   bootblocks[16];         // 0 is bootblock offset, 1-11 sub-sector sizes (or 0)
    cyg_bool     banked;
    cyg_uint32   bufsiz;                 // write buffer size in units of flash_data_t
} flash_dev_info_t;

static const flash_dev_info_t flash_dev_info = {
#include <cyg/io/flash_h8300progs_parts.inl>
};

enum internal_program {none,erase,write};
static enum internal_program download;
static CYG_ADDRWORD lib_top;

void flash_query(void* data);
int  flash_erase_block(void* block, unsigned int size); 
int  flash_program_buf(void* addr, void* data, int len);

static int get_blockno(void *block);
//----------------------------------------------------------------------------
// Auxiliary functions
// Functions that put the flash device into non-read mode must reside
// in RAM.
static int setup_library(enum internal_program type)
    __attribute__ ((section (".2ram.setup_library")));
static int call_library(unsigned long er0,unsigned long er1)
    __attribute__ ((section (".2ram.call_library")));

#define WRITE_SIZE 128
static char flash_write_buffer[WRITE_SIZE]
    __attribute__ ((section (".2ram.flash_write_buffer")));
static char iram_sp[128]
    __attribute__ ((section (".2ram.flash_write_buffer")));

//----------------------------------------------------------------------------
// Flash Query
//
// Not support. dummy.

void
flash_query(void* data)
{
}


//----------------------------------------------------------------------------
// Initialize driver details
int
flash_hwr_init(void)
{
    lib_top = CYGHWR_FLASH_H8300_PLF_INIT();
    if (!lib_top)
	return FLASH_ERR_INVALID;

    // Hard wired for now
    flash_info.block_size = flash_dev_info.block_size;
    flash_info.blocks = flash_dev_info.block_count;
    flash_info.start = (void *)CYGNUM_FLASH_BASE;
    flash_info.end = (void *)(CYGNUM_FLASH_BASE+flash_dev_info.device_size);
    iram_sp[0] = 0;
    return FLASH_ERR_OK;
}

//----------------------------------------------------------------------------
// Map a hardware status to a package error
int
flash_hwr_map_error(int e)
{
    return e;
}


//----------------------------------------------------------------------------
// See if a range of FLASH addresses overlaps currently running code
bool
flash_code_overlaps(void *start, void *end)
{
    extern unsigned char _stext[], _etext[];

    return ((((unsigned long)&_stext >= (unsigned long)start) &&
             ((unsigned long)&_stext < (unsigned long)end)) ||
            (((unsigned long)&_etext >= (unsigned long)start) &&
             ((unsigned long)&_etext < (unsigned long)end)));
}

//----------------------------------------------------------------------------
// Erase Block

int
flash_erase_block(void* block, unsigned int size)
{
    int len = 0;
    int blockno;
    cyg_bool bootblock = false;
    cyg_uint32 *bootblocks = (cyg_uint32 *)0;

    // Assume not "boot" sector, full size
    bootblock = false;
    len = flash_dev_info.block_size;

    // Is this in a "boot" sector?
    if (flash_dev_info.bootblock) {
        bootblocks = (cyg_uint32 *)&flash_dev_info.bootblocks[0];
        while (*bootblocks != _LAST_BOOTBLOCK) {
            if (*bootblocks++ == (unsigned long)block) {
                len = *bootblocks++;  // Size of first sub-block
                bootblock = true;
                break;
            } else {
                int ls = flash_dev_info.block_size;
                // Skip over segment
                while ((ls -= *bootblocks++) > 0) ;
            }
        }
    }

    // setup erase program
    if ((download != erase) && setup_library(erase))
	return FLASH_ERR_HWR;
	
    blockno = get_blockno(block);
    while (size > 0) {
	if ((*(volatile char *)CYGARC_FCCS & 0x90) != 0x80) // write protect / error detect
            return FLASH_ERR_PROTECT;

 	if (call_library(blockno,0))
	    return FLASH_ERR_HWR;
	size -= len;
	blockno++;
	len = flash_dev_info.block_size;
	if (bootblock) {
	    if (*bootblocks != _LAST_BOOTBLOCK)
		len = *bootblocks++;
	    else
		bootblock = false;
	}
    }
    return FLASH_ERR_OK;
}


//----------------------------------------------------------------------------
// Program Buffer
int
flash_program_buf(void* addr, void* data, int len)
{
    int i;
    int write_offset,write_length;
    char *align_write_addr;
    char *buf_p;
    char *data_p = (char *)data;
    bool write_data;

    // setup write program
    if ((download != write) && setup_library(write))
	return FLASH_ERR_HWR;
	
    while (len > 0) {
	align_write_addr = (void *)((unsigned long)addr & ~0x7f); // align 128byte
	write_offset = (unsigned long)addr - (unsigned long)align_write_addr;
	write_length = (len < WRITE_SIZE)?len:WRITE_SIZE;
	buf_p = flash_write_buffer;

	if ((*(volatile char *)CYGARC_FCCS & 0x90) != 0x80) // write protect / error detect
            return FLASH_ERR_PROTECT;

	// setup write data
	write_data = false;
	if (write_offset > 0)
	    for (i = 0; i < write_offset; i++)
		*buf_p++ = 0xff;
	for (i = 0; i < write_length; i++) {
	    write_data = ((*data_p & 0xff) != 0xff);
	    *buf_p++ = *data_p++;
	}
	if (write_length < WRITE_SIZE)
	    for (i = write_length; i < WRITE_SIZE; i++)
		*buf_p++ = 0xff;

	if (write_data) {
	    // writeing
	    if (call_library((unsigned long)flash_write_buffer,(unsigned long)align_write_addr))
		return FLASH_ERR_HWR;
	}
	addr = (void *)((unsigned long)addr + write_length);
	len  -= write_length;
    }
    return FLASH_ERR_OK;
}

static int get_blockno(void *addr)
{
    int blockno = 0;
    cyg_uint32 *bootblocks;
    cyg_uint32 blocktop = 0;

    if (flash_dev_info.bootblock) {
        bootblocks = (cyg_uint32 *)&flash_dev_info.bootblocks[0];
	blocktop = bootblocks[1];
        while (*bootblocks != _LAST_BOOTBLOCK) {
            if (blocktop > (unsigned long)addr)
		return blockno;
            else {
		bootblocks++;
		blocktop += bootblocks[1];
		blockno++;
            }
        }
    }
    return ((cyg_uint32)addr - blocktop + 1) / flash_dev_info.block_size + blockno - 1;
}

static int setup_library(enum internal_program type)
{
    int result = 0;
    *(volatile cyg_uint8 *)CYGARC_FTDAR &= 0x7f;    //clear TDER

    switch(type) {
    case erase:
	*(volatile cyg_uint8 *)CYGARC_FECS |= 0x01; // select erase program
	break ; 
    case write:
	*(volatile cyg_uint8 *)CYGARC_FPCS |= 0x01; // select write program
	break ; 
    default:
	return -1;
    }
    download = type;

#ifndef SIM_DEBUG
    *(volatile cyg_uint8 *)CYGARC_FKEY = 0xa5;      // enable download
    *(volatile cyg_uint8 *)CYGARC_FCCS |= 0x01;     // download
    __asm__("nop\n\tnop\n\tnop\n\tnop");            // delay
    *(volatile cyg_uint8 *)CYGARC_FKEY = 0x00;      // disable download
    
    result = *(cyg_uint8 *)lib_top;                 // download status
    result |= *(volatile cyg_uint8 *)CYGARC_FTDAR & 0x80;
    if (result) {
	download = none;                            // download failed
	return result;
    }
    // download success
    // initialize library
    __asm__("stc ccr,@-sp\n\t"
	    "orc #0x80,ccr\n\t"
	    "mov.l sp,er3\n\t"
	    "mov.l #_iram_sp+128,sp\n\t"
	    "mov.l %1,er0\n\t"
	    "sub.l er1,er1\n\t"
	    "mov.l %2,er2\n\t"
	    "jsr @er2\n\t"
	    "extu.w r0\n\t"
	    "extu.l er0\n\t"
	    "mov.l er0,%0\n\t"
	    "mov.l er3,sp\n\t"
	    "ldc @sp+,ccr"
	    :"=r"(result)
	    :"g"(CYGHWR_HAL_H8300_CPG_INPUT / 10000),"g"(lib_top + 32)
	    :"er0","er1","er2","er3");
    if (result)
	download = none;                            // setup failed
#endif
    return result;
}

#ifndef SIM_DEBUG
static int call_library(unsigned long er0,unsigned long er1)
{
    int result = 0;
    if (download == none)
	return -1;
#if defined(REFRESH_MEM)
    REFRESH_MEM();
#endif
    H8300_NMI_DISABLE();
    *(volatile cyg_uint8 *)CYGARC_FKEY = 0x5a;      // enable writeing
    __asm__("stc ccr,@-sp\n\t"
	    "orc #0x80,ccr\n\t"
	    "mov.l sp,er4\n\t"
	    "mov.l #_iram_sp+128,sp\n\t"
	    "mov.l %1,er0\n\t"
	    "mov.l %2,er1\n\t"
	    "mov.l %3,er2\n\t"
#if defined(PRE_CALL_PROG)
	    PRE_CALL_PROG
#endif
	    "jsr @er2\n\t"
#if defined(POST_CALL_PROG)
	    POST_CALL_PROG
#endif
	    "extu.w r0\n\t"
	    "extu.l er0\n\t"
	    "mov.l er0,%0\n\t"
	    "mov.l er4,sp\n\t"
	    "ldc @sp+,ccr"
	    :"=r"(result)
	    :"g"(er0),"g"(er1),"g"(lib_top + 16)
	    :"er0","er1","er2","er3","er4");
    *(volatile cyg_uint8 *)CYGARC_FKEY = 0x00;      // disable writeing
    H8300_NMI_ENABLE();
#if defined(REFRESH_MEM)
    REFRESH_MEM();
#endif
    return result;
}
#else
#include <stdlib.h>
static int call_library(unsigned long er0,unsigned long er1)
{
    switch (download) {
    case erase: {
	if (er0 < 8)
	    memset((unsigned char *)(0x1000 * er0),0xff,0x1000);
	else if (er0 == 8)
	    memset((unsigned char *)0x8000,0xff,0x8000);
	else if (er0 < 16)
	    memset((unsigned char *)(0x10000 * (er0 - 8)),0xff,0x10000);
	else
	    return -1;
	return 0;
    }
    case write:
	memcpy((unsigned char *)er1,(unsigned char *)er0,128);
	return 0;
    default:
	return -1;
    }
}
#endif

#endif // CYGONCE_DEVS_FLASH_H8300_PROGS_INL
